from functools import lru_cache
import sys
import os
from pyarrow import fs
from dataclasses import dataclass
from typing import Any, Iterator, List, Optional
from skyplane.exceptions import NoSuchObjectException
from skyplane.obj_store.object_store_interface import ObjectStoreInterface, ObjectStoreObject
from skyplane.utils import logger
import mimetypes


def test_and_set_hadoop_classpath():
    import subprocess

    if "hadoop" in os.environ.get("CLASSPATH", ""):
        return

    # If HADOOP_HOME is not set, set it to the default location
    if "HADOOP_HOME" not in os.environ:
        os.environ["HADOOP_HOME"] = "/usr/local/hadoop"
        os.environ["JAVA_HOME"] = "/usr/lib/jvm/java-11-openjdk-amd64"

    hadoop_bin = os.path.normpath(os.environ["HADOOP_HOME"]) + "/bin/"  #'{0}/bin/hadoop'.format(os.environ['HADOOP_HOME'])
    hadoop_bin_exe = os.path.join(hadoop_bin, "hadoop")

    classpath = subprocess.check_output([hadoop_bin_exe, "classpath", "--glob"])
    os.environ["CLASSPATH"] = classpath.decode("utf-8")


def resolve_hostnames():
    if os.path.exists("/tmp/hostname"):
        logger.info("Found hostname file")
        rc = os.system("cat /tmp/hostname >> /etc/hosts")
        if rc:
            logger.info("Failed to add hostname to /etc/hosts")
    # elif os.path.exists("scripts/on_prem/hostname"):
    #     print("Found hostname file")
    #     os.system("cat scripts/on_prem/hostname >> sudo /etc/hosts")


@dataclass
class HDFSFile(ObjectStoreObject):
    def full_path(self):
        return f"hdfs://{self.key}"


class HDFSInterface(ObjectStoreInterface):
    def __init__(self, host, path="", port=8020):
        self.host = host
        self.port = port
        self.hdfs_path = host
        test_and_set_hadoop_classpath()
        resolve_hostnames()
        self.hdfs = fs.HadoopFileSystem(
            host=f"{self.host}/",
            port=self.port,
            user="hadoop",
            extra_conf={
                "dfs.permissions.enabled": "false",
                "dfs.client.use.datanode.hostname": "true",
                "dfs.datanode.use.datanode.hostname": "false",
            },
        )
        logger.info(f"Connecting to HDFS at {self.host}:{self.port}", flush=True)

    def path(self) -> str:
        return self.hdfs_path

    def list_objects(self, prefix="/skyplane5") -> Iterator[HDFSFile]:
        fileselector = fs.FileSelector("/skyplane5", recursive=True, allow_not_found=True)
        logger.info(f"File selector created successfully, {fileselector.base_dir}")
        response = self.hdfs.get_file_info(fileselector)
        logger.info(f"Response: {response}")
        if hasattr(response, "__len__") and (not isinstance(response, str)):
            for file in response:
                yield HDFSFile(provider="hdfs", bucket=self.host, key=file.path, size=file.size, last_modified=file.mtime)
        else:
            yield HDFSFile(provider="hdfs", bucket=self.host, key=response.path, size=response.size, last_modified=response.mtime)

    def exists(self, obj_name: str):
        try:
            self.get_obj_metadata(obj_name)
            return True
        except NoSuchObjectException:
            return False

    def region_tag(self) -> str:
        return "hdfs:us-central1-a"

    def bucket(self) -> str:
        return self.hdfs_path

    def create_bucket(self, region_tag: str):
        self.hdfs.create_dir("/skyplane5")

    def delete_bucket(self):
        self.hdfs.delete_dir("/skyplane5")

    def bucket_exists(self) -> bool:
        return True

    def get_obj_size(self, obj_name) -> int:
        return self.get_obj_metadata(obj_name).size

    def get_obj_last_modified(self, obj_name):
        return self.get_obj_metadata(obj_name).mtime

    def get_obj_mime_type(self, obj_name):
        return mimetypes.guess_type(obj_name)[0]

    def delete_objects(self, keys: List[str]):
        for key in keys:
            self.hdfs.delete_file(f"/skyplane5/{key}")
        return True

    def download_object(
        self, src_object_name, dst_file_path, offset_bytes=None, size_bytes=None, write_at_offset=False, generate_md5: bool = False
    ):
        with self.hdfs.open_input_stream(f"{src_object_name}") as f1:
            with open(dst_file_path, "wb+" if write_at_offset else "wb") as f2:
                b = f1.read(nbytes=size_bytes)
                while b:
                    f2.write(b)
                    b = f1.read(nbytes=size_bytes)
        return self.get_obj_mime_type(src_object_name), None

    def upload_object(
        self,
        src_file_path,
        dst_object_name,
        part_number=None,
        upload_id=None,
        check_md5: Optional[bytes] = None,
        mime_type: Optional[str] = None,
    ):
        with open(src_file_path, "rb") as f1:
            with self.hdfs.open_output_stream(f"{dst_object_name}") as f2:
                b = f1.read()
                f2.write(b)

    def read_file(self, file_name, offset=0, length=sys.maxsize):
        with self.hdfs.open_input_stream(file_name) as f:
            return print(f.readall())

    def write_file(self, file_name, data, offset=0):
        with self.hdfs.open_output_stream(file_name) as f:
            f.write(data)

    def initiate_multipart_upload(self, dst_object_name: str, mime_type: Optional[str] = None) -> str:
        raise NotImplementedError(f"Multipart upload is not supported for the POSIX file system.")

    def complete_multipart_upload(self, dst_object_name: str, upload_id: str, metadata: Optional[Any] = None) -> None:
        raise NotImplementedError(f"Multipart upload is not supported for the POSIX file system.")

    @lru_cache(maxsize=1024)
    def get_obj_metadata(self, obj_name) -> fs.FileInfo:
        response = self.hdfs.get_file_info(obj_name)
        if response.type is fs.FileType.NotFound:
            raise NoSuchObjectException(f"Object {obj_name} not found")
        else:
            return response



import functools
import shutil
from collections import namedtuple
from dataclasses import dataclass

import numpy as np
import pandas as pd
from typing import Dict, List, Optional, Tuple

from skyplane import compute
from skyplane.planner.topology import TopologyPlan
from skyplane.utils import logger
from skyplane.utils.definitions import GB

GBIT_PER_GBYTE = 8


@dataclass
class ThroughputProblem:
    src: str
    dst: str
    required_throughput_gbits: float
    gbyte_to_transfer: float
    instance_limit: int
    const_throughput_grid_gbits: Optional[np.ndarray] = None  # if not set, load from profiles
    const_cost_per_gb_grid: Optional[np.ndarray] = None  # if not set, load from profiles
    # provider bandwidth limits (egress, ingress)
    aws_instance_throughput_limit: Tuple[float, float] = (5, 10)
    gcp_instance_throughput_limit: Tuple[float, float] = (7, 16)  # limited to 12.5 gbps due to CPU limit
    azure_instance_throughput_limit: Tuple[float, float] = (16, 16)  # limited to 12.5 gbps due to CPU limit
    # benchmarked_throughput_connections is the number of connections that the iperf3 throughput grid was run at,
    # we assume throughput is linear up to this connection limit
    benchmarked_throughput_connections = 64
    cost_per_instance_hr = 0.54  # based on m5.8xlarge spot
    instance_cost_multiplier = 1.0
    instance_provision_time_s = 0.0

    def to_summary_dict(self):
        """Simple summary of the problem"""
        return {
            "src": self.src,
            "dst": self.dst,
            "required_throughput_gbits": self.required_throughput_gbits,
            "gbyte_to_transfer": self.gbyte_to_transfer,
            "instance_limit": self.instance_limit,
            "aws_instance_throughput_limit": self.aws_instance_throughput_limit,
            "gcp_instance_throughput_limit": self.gcp_instance_throughput_limit,
            "azure_instance_throughput_limit": self.azure_instance_throughput_limit,
            "benchmarked_throughput_connections": self.benchmarked_throughput_connections,
            "cost_per_instance_hr": self.cost_per_instance_hr,
            "instance_cost_multiplier": self.instance_cost_multiplier,
            "instance_provision_time_s": self.instance_provision_time_s,
        }


@dataclass
class ThroughputSolution:
    problem: ThroughputProblem
    is_feasible: bool
    extra_data: Optional[Dict] = None

    # solution variables
    var_edge_flow_gigabits: Optional[np.ndarray] = None
    var_conn: Optional[np.ndarray] = None
    var_instances_per_region: Optional[np.ndarray] = None

    # solution values
    throughput_achieved_gbits: Optional[List[float]] = None
    cost_egress_by_edge: Optional[np.ndarray] = None
    cost_egress: Optional[float] = None
    cost_instance: Optional[float] = None
    cost_total: Optional[float] = None
    transfer_runtime_s: Optional[float] = None

    # baseline
    baseline_throughput_achieved_gbits: Optional[float] = None
    baseline_cost_egress: Optional[float] = None
    baseline_cost_instance: Optional[float] = None
    baseline_cost_total: Optional[float] = None

    def to_summary_dict(self):
        """Print simple summary of solution."""
        if self.is_feasible:
            return {
                "is_feasible": self.is_feasible,
                "solution": {
                    "throughput_achieved_gbits": self.throughput_achieved_gbits,
                    "cost_egress": self.cost_egress,
                    "cost_instance": self.cost_instance,
                    "cost_total": self.cost_total,
                    "transfer_runtime_s": self.transfer_runtime_s,
                },
                "baseline": {
                    "throughput_achieved_gbits": self.baseline_throughput_achieved_gbits,
                    "cost_egress": self.baseline_cost_egress,
                    "cost_instance": self.baseline_cost_instance,
                    "cost_total": self.baseline_cost_total,
                },
            }
        else:
            return {"is_feasible": self.is_feasible}


class ThroughputSolver:
    def __init__(self, df_path, default_throughput=0.0):
        self.df = pd.read_csv(df_path).set_index(["src_region", "dst_region", "src_tier", "dst_tier"]).sort_index()
        self.default_throughput = default_throughput

    @functools.lru_cache(maxsize=None)
    def get_path_throughput(self, src, dst, src_tier="PREMIUM", dst_tier="PREMIUM"):
        if src == dst:
            return self.default_throughput
        elif (src, dst, src_tier, dst_tier) not in self.df.index:
            return None
        return self.df.loc[(src, dst, src_tier, dst_tier), "throughput_sent"].values[0]

    @functools.lru_cache(maxsize=None)
    def get_path_cost(self, src, dst, src_tier="PREMIUM", dst_tier="PREMIUM"):
        assert src_tier == "PREMIUM" and dst_tier == "PREMIUM"
        return compute.CloudProvider.get_transfer_cost(src, dst)

    def get_regions(self):
        return list(sorted(set(list(self.df.index.levels[0].unique()) + list(self.df.index.levels[1].unique()))))

    def get_throughput_grid(self):
        regions = self.get_regions()
        data_grid = np.zeros((len(regions), len(regions)))
        for i, src in enumerate(regions):
            for j, dst in enumerate(regions):
                data_grid[i, j] = self.get_path_throughput(src, dst) if self.get_path_throughput(src, dst) is not None else 0
        data_grid = data_grid / GB
        return data_grid.round(4)

    def get_cost_grid(self):
        regions = self.get_regions()
        data_grid = np.zeros((len(regions), len(regions)))
        for i, src in enumerate(regions):
            for j, dst in enumerate(regions):
                cost = self.get_path_cost(src, dst)
                assert cost is not None and cost >= 0, f"Cost for {src} -> {dst} is {cost}"
                data_grid[i, j] = cost
        return data_grid.round(2)

    def get_baseline_throughput_and_cost(self, p: ThroughputProblem) -> Tuple[float, float, float]:
        src, dst = p.src, p.dst
        throughput = max(p.instance_limit * self.get_path_throughput(src, dst) / GB, 1e-6)
        transfer_s = p.gbyte_to_transfer * GBIT_PER_GBYTE / throughput
        instance_cost = p.cost_per_instance_hr * p.instance_limit * transfer_s / 3600
        egress_cost = p.gbyte_to_transfer * self.get_path_cost(src, dst)
        return throughput, egress_cost, instance_cost

    def plot_throughput_grid(self, data_grid, title="Throughput (Gbps)"):
        import matplotlib.pyplot as plt  # pytype: disable=import-error

        for i in range(data_grid.shape[0]):
            for j in range(data_grid.shape[1]):
                if i <= j:
                    data_grid[i, j] = np.nan

        regions = self.get_regions()
        fig, ax = plt.subplots(1, 1, figsize=(9, 9))
        ax.imshow(data_grid)
        ax.set_title(title)
        ax.set_xticks(np.arange(len(regions)))
        ax.set_yticks(np.arange(len(regions)))
        ax.set_xticklabels(regions)
        ax.set_yticklabels(regions)

        for tick in ax.get_xticklabels():
            tick.set_rotation(90)

        # compute mean point of non nan values
        mean_point = np.nanmean(data_grid)
        for i, row in enumerate(data_grid):
            for j, col in enumerate(row):
                if i > j:
                    ax.text(j, i, round(col, 1), ha="center", va="center", color="white" if col < mean_point else "black")

        fig.patch.set_facecolor("white")
        fig.subplots_adjust(hspace=0.6)
        ax.figure.colorbar(ax.images[0], ax=ax)
        return fig, ax

    def print_solution(self, solution: ThroughputSolution):
        if solution.is_feasible:
            regions = self.get_regions()
            logger.debug(
                f"Total cost: ${solution.cost_total:.4f} (egress: ${solution.cost_egress:.4f}, instance: ${solution.cost_instance:.4f})"
            )
            logger.debug(f"Total throughput: [{', '.join(str(round(t, 2)) for t in solution.throughput_achieved_gbits)}] Gbps")
            logger.debug(f"Total runtime: {solution.transfer_runtime_s:.2f}s")
            region_inst_count = {regions[i]: int(solution.var_instances_per_region[i]) for i in range(len(regions))}
            logger.debug("Instance regions: [{}]".format(", ".join(f"{r}={c}" for r, c in region_inst_count.items() if c > 0)))
            logger.debug("Flow matrix:")
            for i, src in enumerate(regions):
                for j, dst in enumerate(regions):
                    if solution.var_edge_flow_gigabits[i, j] > 0:
                        gb_sent = solution.transfer_runtime_s * solution.var_edge_flow_gigabits[i, j] * GBIT_PER_GBYTE
                        s = f"\t{src} -> {dst}: {solution.var_edge_flow_gigabits[i, j]:.2f} Gbps with {solution.var_conn[i, j]:.1f} connections, "
                        s += f"{gb_sent:.1f}GB (link capacity = {solution.problem.const_throughput_grid_gbits[i, j]:.2f} Gbps)"
                        logger.debug(s)
        else:
            logger.debug("No feasible solution")

    def plot_graphviz(self, solution: ThroughputSolution):
        import graphviz as gv  # pytype: disable=import-error

        # if dot is not installed
        has_dot = shutil.which("dot") is not None
        if not has_dot:
            logger.error("Graphviz is not installed. Please install it to plot the solution (sudo apt install graphviz).")
            return None

        regions = self.get_regions()
        g = gv.Digraph(name="throughput_graph")
        g.attr(rankdir="LR")

        label = f"{solution.problem.src} to {solution.problem.dst}\n"
        label += f"[{', '.join(str(round(t, 2)) for t in solution.throughput_achieved_gbits)}] Gbps, ${solution.cost_total:.4f}\n"
        label += f"Average: ${solution.cost_total / solution.transfer_runtime_s:.4f}/GB"
        g.attr(label=label)
        g.attr(labelloc="t")
        for i, src in enumerate(regions):
            for j, dst in enumerate(regions):
                if solution.var_edge_flow_gigabits[i, j] > 0:
                    link_cost = self.get_path_cost(src, dst)
                    label = f"{solution.var_edge_flow_gigabits[i, j]:.2f} Gbps (of {solution.problem.const_throughput_grid_gbits[i, j]:.2f}Gbps), "
                    label += f"\n${link_cost:.4f}/GB over {solution.var_conn[i, j]:.1f}c"
                    src_label = f"{src.replace(':', '/')}x{solution.var_instances_per_region[i]:.1f}"
                    dst_label = f"{dst.replace(':', '/')}x{solution.var_instances_per_region[j]:.1f}"
                    g.edge(src_label, dst_label, label=label)
        return g

    def to_replication_topology(self, solution: ThroughputSolution, scale_to_capacity=True) -> Tuple[TopologyPlan, float]:
        regions = self.get_regions()
        Edge = namedtuple("Edge", ["src_region", "src_instance_idx", "dst_region", "dst_instance_idx", "connections"])

        # check is feasible
        if not solution.is_feasible:
            raise ValueError("Solution is not feasible")

        # compute connections to target per instance
        ninst = solution.var_instances_per_region
        average_egress_conns = [0 if ninst[i] == 0 else np.ceil(solution.var_conn[i, :].sum() / ninst[i]) for i in range(len(regions))]
        average_ingress_conns = [0 if ninst[i] == 0 else np.ceil(solution.var_conn[:, i].sum() / ninst[i]) for i in range(len(regions))]

        # first assign source instances to destination regions
        src_edges: List[Edge] = []
        n_instances = {}
        for i, src in enumerate(regions):
            src_instance_idx, src_instance_connections = 0, 0
            for j, dst in enumerate(regions):
                if solution.var_edge_flow_gigabits[i, j] > 0:
                    connections_to_allocate = np.rint(solution.var_conn[i, j]).astype(int)
                    while connections_to_allocate > 0:
                        # if this edge would exceed the instance connection limit, partially add connections to
                        # current instance and increment instance
                        if connections_to_allocate + src_instance_connections > average_egress_conns[i]:
                            partial_conn = average_egress_conns[i] - src_instance_connections
                            connections_to_allocate -= partial_conn
                            assert connections_to_allocate >= 0, f"connections_to_allocate = {connections_to_allocate}"
                            assert partial_conn >= 0, f"partial_conn = {partial_conn}"
                            if partial_conn > 0:
                                src_edges.append(Edge(src, src_instance_idx, dst, None, partial_conn))
                                logger.fs.warning(
                                    f"{src}:{src_instance_idx}:{src_instance_connections}c -> {dst} (partial): {partial_conn}c of {connections_to_allocate}c remaining"
                                )
                            src_instance_idx += 1
                            src_instance_connections = 0
                        else:
                            partial_conn = connections_to_allocate
                            connections_to_allocate = 0
                            src_edges.append(Edge(src, src_instance_idx, dst, None, partial_conn))
                            logger.fs.warning(
                                f"{src}:{src_instance_idx}:{src_instance_connections}c -> {dst}: {partial_conn}c of {connections_to_allocate}c remaining"
                            )
                            src_instance_connections += partial_conn
            n_instances[i] = src_instance_idx + 1

        # assign destination instances (currently None) to Edges
        dst_edges = []
        dsts_instance_idx = {i: 0 for i in regions}
        dsts_instance_conn = {i: 0 for i in regions}
        for e in src_edges:
            connections_to_allocate = np.rint(e.connections).astype(int)
            while connections_to_allocate > 0:
                if connections_to_allocate + dsts_instance_conn[e.dst_region] > average_ingress_conns[regions.index(e.dst_region)]:
                    partial_conn = average_ingress_conns[regions.index(e.dst_region)] - dsts_instance_conn[e.dst_region]
                    connections_to_allocate = connections_to_allocate - partial_conn
                    if partial_conn > 0:
                        dst_edges.append(
                            Edge(e.src_region, e.src_instance_idx, e.dst_region, dsts_instance_idx[e.dst_region], partial_conn)
                        )
                        logger.fs.warning(
                            f"{e.src_region}:{e.src_instance_idx}:{dsts_instance_conn[e.dst_region]}c -> {e.dst_region}:{dsts_instance_idx[e.dst_region]}:{dsts_instance_conn[e.dst_region]}c (partial): {partial_conn}c of {connections_to_allocate}c remaining"
                        )
                    dsts_instance_idx[e.dst_region] += 1
                    dsts_instance_conn[e.dst_region] = 0
                else:
                    dst_edges.append(
                        Edge(e.src_region, e.src_instance_idx, e.dst_region, dsts_instance_idx[e.dst_region], connections_to_allocate)
                    )
                    logger.fs.warning(
                        f"{e.src_region}:{e.src_instance_idx}:{dsts_instance_conn[e.dst_region]}c -> {e.dst_region}:{dsts_instance_idx[e.dst_region]}:{dsts_instance_conn[e.dst_region]}c: {connections_to_allocate}c remaining"
                    )
                    dsts_instance_conn[e.dst_region] += connections_to_allocate
                    connections_to_allocate = 0

        # scale connections up to saturate links
        if scale_to_capacity:
            conns_egress: Dict[Tuple[str, int], int] = {}
            conns_ingress: Dict[Tuple[str, int], int] = {}
            for e in dst_edges:
                conns_egress[(e.src_region, e.src_instance_idx)] = e.connections + conns_egress.get((e.src_region, e.src_instance_idx), 0)
                conns_ingress[(e.dst_region, e.dst_instance_idx)] = e.connections + conns_ingress.get((e.dst_region, e.dst_instance_idx), 0)
            bottleneck_capacity = max(list(conns_egress.values()) + list(conns_ingress.values()))
            scale_factor = 64 / bottleneck_capacity
            logger.fs.warning(f"Scaling connections by {scale_factor:.2f}x")
            dst_edges = [e._replace(connections=int(e.connections * scale_factor)) for e in dst_edges]
        else:
            scale_factor = 1.0

        # build TopologyPlan
        obj_store_edges = set()
        replication_topology = TopologyPlan(src_region_tag=e.src_region, dest_region_tags=[e.dst_region])
        for e in dst_edges:
            if e.connections >= 1:
                # connect source to destination
                replication_topology.add_instance_instance_edge(
                    src_region=e.src_region,
                    src_instance=e.src_instance_idx,
                    dest_region=e.dst_region,
                    dest_instance=e.dst_instance_idx,
                    num_connections=e.connections,
                )

                # connect source instances to source gateway
                if e.src_region == solution.problem.src and ("src", e.src_region, e.src_instance_idx) not in obj_store_edges:
                    replication_topology.add_objstore_instance_edge(
                        src_region=e.src_region, dest_region=e.src_region, dest_instance=e.src_instance_idx
                    )
                    obj_store_edges.add(("src", e.src_region, e.src_instance_idx))

                # connect destination instances to destination gateway
                if e.dst_region == solution.problem.dst and ("dst", e.dst_region, e.dst_instance_idx) not in obj_store_edges:
                    replication_topology.add_instance_objstore_edge(
                        src_region=e.dst_region, src_instance=e.dst_instance_idx, dest_region=e.dst_region
                    )
                    obj_store_edges.add(("dst", e.dst_region, e.dst_instance_idx))

        return replication_topology, scale_factor



# (C) Copyright Samsung SDS. 2023

#

# Licensed under the Apache License, Version 2.0 (the "License");

# you may not use this file except in compliance with the License.

# You may obtain a copy of the License at

#

#     http://www.apache.org/licenses/LICENSE-2.0

#

# Unless required by applicable law or agreed to in writing, software

# distributed under the License is distributed on an "AS IS" BASIS,

# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

# See the License for the specific language governing permissions and

# limitations under the License.
""" SCP Open-API utility functions """

import base64
import datetime
from functools import wraps
import hashlib
import hmac
import os
import json
import random
import time
from urllib import parse
import requests
from skyplane.utils import logger

CREDENTIALS_PATH = "~/.scp/scp_credential"
API_ENDPOINT = "https://openapi.samsungsdscloud.com"
TEMP_VM_JSON_PATH = "/tmp/json/tmp_vm_body.json"


class SCPClientError(Exception):
    pass


class SCPOngoingRequestError(Exception):
    pass


class SCPCreationFailError(Exception):
    pass


def raise_scp_error(response: requests.Response) -> None:
    """Raise SCPCloudError if appropriate."""
    status_code = response.status_code
    if status_code == 200 or status_code == 202:
        return
    try:
        resp_json = response.json()
        message = resp_json["message"]
    except (KeyError, json.decoder.JSONDecodeError):
        # print(f"response: {response.content}")
        raise SCPClientError(f"Unexpected error. Status code: {status_code}")

    raise SCPClientError(f"{status_code}: {message}")


def _retry(method, max_tries=60, backoff_s=1):
    @wraps(method)
    def method_with_retries(self, *args, **kwargs):
        try_count = 0
        while try_count < max_tries:
            try:
                return method(self, *args, **kwargs)
            except Exception as e:
                # print(e.args[0])
                retry_codes = ["500", "412", "403"]  # add 403 for object storage
                # if any(code in e.args[0] for code in retry_codes):
                if any(code in str(e) for code in retry_codes):
                    try_count += 1
                    # console.print(f"[yellow] retries: {method.__name__} - {e}, try_count : {try_count}[/yellow]")
                    logger.fs.debug(f"retries: {method.__name__} - {e}, try_count : {try_count}")
                    if try_count < max_tries:
                        time.sleep(backoff_s)
                    else:
                        raise e
                elif "Connection aborted" in str(e):
                    try_count += 1
                    # with open(f"/skyplane/retry_nocode_error.txt", "w") as f:
                    #     f.write(str(e))
                    logger.fs.debug(f"retries: {method.__name__} - {e}, try_count : {try_count}")
                    if try_count < max_tries:
                        time.sleep(backoff_s)
                    else:
                        raise e
                else:
                    # with open(f"/skyplane/retry_nocode_error.txt", "w") as f:
                    #     f.write(str(e))
                    raise e

    return method_with_retries


class SCPClient:
    """SCP Open-API client"""

    def __init__(self) -> None:
        # print('SCPClient init')
        self.credentials = os.path.expanduser(CREDENTIALS_PATH)
        if not os.path.exists(self.credentials):
            self.credentials = os.path.expanduser("/pkg/data/scp_credential")
        assert os.path.exists(self.credentials), "SCP Credentials not found"
        with open(self.credentials, "r") as f:
            lines = [line.strip() for line in f.readlines() if " = " in line]
            self._credentials = {line.split(" = ")[0]: line.split(" = ")[1] for line in lines}

        self.access_key = self._credentials["scp_access_key"]
        self.secret_key = self._credentials["scp_secret_key"]
        self.project_id = self._credentials["scp_project_id"]
        self.client_type = "OpenApi"
        self.timestamp = ""
        self.signature = ""

        self.headers = {
            "X-Cmp-AccessKey": f"{self.access_key}",
            "X-Cmp-ClientType": f"{self.client_type}",
            "X-Cmp-ProjectId": f"{self.project_id}",
            "X-Cmp-Timestamp": f"{self.timestamp}",
            "X-Cmp-Signature": f"{self.signature}",
        }

    @_retry
    def _get(self, url, contents_key="contents"):
        method = "GET"
        url = f"{API_ENDPOINT}{url}"
        self.set_timestamp()
        self.set_signature(url=url, method=method)

        response = requests.get(url, headers=self.headers)
        raise_scp_error(response)

        if contents_key is not None:
            return response.json().get(contents_key, [])
        else:
            return response.json()

    @_retry
    def _getDetail(self, url):
        method = "GET"
        url = f"{API_ENDPOINT}{url}"
        self.set_timestamp()
        self.set_signature(url=url, method=method)

        response = requests.get(url, headers=self.headers)
        response_detail = response.content.decode("utf-8")
        raise_scp_error(response)
        return json.loads(response_detail)

    def post(self, url, req_body):
        method = "POST"
        url = f"{API_ENDPOINT}{url}"
        self.set_timestamp()
        self.set_signature(url=url, method=method)

        response = requests.post(url, json=req_body, headers=self.headers)
        raise_scp_error(response)
        return response.json()

    @_retry
    def _post(self, url, req_body):
        return self.post(url, req_body)

    @_retry
    def _delete(self, url, req_body=None):
        method = "DELETE"
        url = f"{API_ENDPOINT}{url}"
        self.set_timestamp()
        self.set_signature(url=url, method=method)

        # try:
        if req_body:
            response = requests.delete(url, json=req_body, headers=self.headers)
        else:
            response = requests.delete(url, headers=self.headers)
        raise_scp_error(response)
        try:
            return response.json()
        except json.JSONDecodeError as e:
            # return response.content.decode('utf-8')
            return response

    def wait_for_completion(self, task_name, completion_condition, retry_limit=5, timeout=180):
        start = time.time()
        # console.print(f"[bright_black] Waiting for... {task_name} [/bright_black]")
        logger.fs.debug(f"Waiting for... {task_name}")

        retries = 0
        while time.time() - start < timeout and retries < retry_limit:
            try:
                if completion_condition():
                    break
            except Exception as e:
                retries += 1
                # print(f"Exception occurred during completion_condition(): {task_name} - {e}, retries : {retries}")
                logger.fs.debug(f"Exception occurred during completion_condition(): {task_name} - {e}, retries : {retries}")
            time.sleep(1)
        if retries == retry_limit:
            # console.print(f"[red] {task_name}... Retry limit reached [/red]")
            logger.fs.error(f"{task_name}... Retry limit reached")
            raise SCPCreationFailError(f"Failed to create {task_name}")
        else:
            # console.print(f"[green] {task_name}... Completed [/green]")
            logger.fs.debug(f"{task_name}... Completed")

    def set_timestamp(self) -> None:
        # self.timestamp = str(int(round((datetime.datetime.now() - datetime.timedelta(minutes=1)).timestamp() * 1000)))
        current_time = datetime.datetime.now()
        self.timestamp = str(int(round(current_time.timestamp() * 1000)))
        self.headers["X-Cmp-Timestamp"] = self.timestamp

    def set_signature(self, url: str, method: str) -> None:
        self.signature = self.get_signature(url=url, method=method)
        self.headers["X-Cmp-Signature"] = f"{self.signature}"

    def get_signature(self, url: str, method: str) -> str:
        url_info = parse.urlsplit(url)
        url = f"{url_info.scheme}://{url_info.netloc}{parse.quote(url_info.path)}"

        if url_info.query:
            enc_params = [(item[0], parse.quote(item[1][0])) for item in parse.parse_qs(url_info.query).items()]
            url = f"{url}?{parse.urlencode(enc_params)}"

        message = method + url + self.timestamp + self.access_key + self.project_id + self.client_type
        message = message.encode("utf-8")
        secret = self.secret_key.encode("utf-8")

        signature = base64.b64encode(hmac.new(secret, message, digestmod=hashlib.sha256).digest()).decode("utf-8")

        return signature

    def random_wait(self):
        random_int = random.randint(1, 100)
        time.sleep(1 * random_int / 100)



import subprocess
from shlex import split
from typing import Optional

import typer
from rich.prompt import IntPrompt

import skyplane.cli.cli_cloud
import skyplane.cli.cli_config

# import skyplane.cli.experiments # disable experiments
from skyplane import compute

from skyplane.cli.cli_init import init
from skyplane.cli.cli_transfer import cp, sync
from skyplane.cli.impl.common import query_instances
from skyplane.utils import logger
from skyplane.utils.fn import do_parallel

app = typer.Typer(name="skyplane")
app.command(
    name="cp",
    help="Copy files between any two cloud object stores",
)(cp)
app.command(
    name="sync",
    help="Sync files between any two cloud object stores",
)(sync)
app.command(
    name="init",
    help="Initialize the Skyplane CLI with your cloud credentials",
)(init)
# app.add_typer(skyplane.cli.experiments.app, name="experiments") # disable experiments
app.add_typer(skyplane.cli.cli_cloud.app, name="cloud")
app.add_typer(skyplane.cli.cli_config.app, name="config")


@app.command()
def deprovision(
    all: bool = typer.Option(False, "--all", "-a", help="Deprovision all resources including networks."),
    filter_client_id: Optional[str] = typer.Option(None, help="Only deprovision instances with this client ID under the instance tag."),
):
    """Deprovision all resources created by skyplane."""
    instances = query_instances()
    if filter_client_id:
        instances = [instance for instance in instances if instance.tags().get("skyplaneclientid") == filter_client_id]

    if instances:
        typer.secho(f"Deprovisioning {len(instances)} instances", fg="yellow", bold=True)
        do_parallel(lambda instance: instance.terminate_instance(), instances, desc="Deprovisioning", spinner=True, spinner_persist=True)
        # if compute.SCPAuthentication().enabled():
        if any([instance.region_tag.split(":")[0] == "scp" for instance in instances]):
            typer.secho(f"Removing SCP Gateway Rules", fg="yellow", bold=True)
            scp_instances = [instance for instance in instances if instance.region_tag.split(":")[0] == "scp"]
            scp = compute.SCPCloudProvider()
            scp.remove_gateway_rule_all(scp_instances)
    else:
        typer.secho("No instances to deprovision", fg="yellow", bold=True)

    if all:
        if compute.AWSAuthentication().enabled():
            aws = compute.AWSCloudProvider()
            aws.teardown_global()
        if compute.GCPAuthentication().enabled():
            gcp = compute.GCPCloudProvider()
            gcp.teardown_global()
        if compute.AzureAuthentication().enabled():
            azure = compute.AzureCloudProvider()
            azure.teardown_global()
        if compute.SCPAuthentication().enabled():
            scp = compute.SCPCloudProvider()
            scp.teardown_global()


@app.command()
def ssh():
    """SSH into a running gateway."""
    instances = query_instances()
    if len(instances) == 0:
        typer.secho(f"No instances found", fg="red", err=True)
        raise typer.Abort()

    instance_map = {f"{i.region_tag}, {i.public_ip()} ({i.instance_state()})": i for i in instances}
    choices = list(sorted(instance_map.keys()))

    # ask for selection
    typer.secho("Select an instance:", fg="yellow", bold=True)
    for i, choice in enumerate(choices):
        typer.secho(f"{i+1}) {choice}", fg="yellow")
    choice = IntPrompt.ask("Enter an instance number", choices=list([str(i) for i in range(1, len(choices) + 1)]), show_choices=False)
    instance = instance_map[choices[choice - 1]]

    # ssh
    cmd = instance.get_ssh_cmd()
    logger.info(f"Running SSH command: {cmd}")
    logger.info("It may ask for a private key password, try `skyplane`.")
    proc = subprocess.Popen(split(cmd))
    proc.wait()


typer_click_object = typer.main.get_command(app)

if __name__ == "__main__":
    app()
